@@ -8,6 +8,7 @@
 # Distributed under the terms of the Modified BSD License.
 
 from collections import defaultdict
+from datetime import datetime, timedelta
 from functools import partial
 import os
 
@@ -17,14 +18,14 @@ from tornado.ioloop import IOLoop, PeriodicCallback
 
 from jupyter_client.session import Session
 from jupyter_client.multikernelmanager import MultiKernelManager
-from traitlets import Any, Bool, Dict, List, Unicode, TraitError, Integer, default, validate
+from traitlets import (Any, Bool, Dict, List, Unicode, TraitError, Integer,
+       Float, Instance, default, validate
+)
 
 from notebook.utils import to_os_path, exists
 from notebook._tz import utcnow, isoformat
 from ipython_genutils.py3compat import getcwd
 
-from datetime import timedelta
-
 
 class MappingKernelManager(MultiKernelManager):
     """A KernelManager that handles notebook mapping and HTTP error handling"""
@@ -61,11 +62,10 @@ class MappingKernelManager(MultiKernelManager):
             raise TraitError("kernel root dir %r is not a directory" % value)
         return value
 
-    cull_idle_timeout_minimum = 300 # 5 minutes
     cull_idle_timeout = Integer(0, config=True,
-        help="""Timeout (in seconds) after which a kernel is considered idle and ready to be culled.  Values of 0 or
-        lower disable culling. The minimum timeout is 300 seconds (5 minutes). Positive values less than the minimum value
-        will be set to the minimum."""
+        help="""Timeout (in seconds) after which a kernel is considered idle and ready to be culled.
+        Values of 0 or lower disable culling. Very short timeouts may result in kernels being culled
+        for users with poor network connections."""
     )
 
     cull_interval_default = 300 # 5 minutes
@@ -75,12 +75,35 @@ class MappingKernelManager(MultiKernelManager):
 
     cull_connected = Bool(False, config=True,
         help="""Whether to consider culling kernels which have one or more connections.
-        Only effective if cull_idle_timeout is not 0."""
+        Only effective if cull_idle_timeout > 0."""
     )
 
     cull_busy = Bool(False, config=True,
         help="""Whether to consider culling kernels which are busy.
-        Only effective if cull_idle_timeout is not 0."""
+        Only effective if cull_idle_timeout > 0."""
+    )
+
+    buffer_offline_messages = Bool(True, config=True,
+        help="""Whether messages from kernels whose frontends have disconnected should be buffered in-memory.
+
+        When True (default), messages are buffered and replayed on reconnect,
+        avoiding lost messages due to interrupted connectivity.
+
+        Disable if long-running kernels will produce too much output while
+        no frontends are connected.
+        """
+    )
+    
+    kernel_info_timeout = Float(60, config=True,
+        help="""Timeout for giving up on a kernel (in seconds).
+
+        On starting and restarting kernels, we check whether the
+        kernel is running and responsive by sending kernel_info_requests.
+        This sets the timeout in seconds for how long the kernel can take
+        before being presumed dead. 
+        This affects the MappingKernelManager (which handles kernel restarts) 
+        and the ZMQChannelsHandler (which handles the startup).
+        """
     )
 
     _kernel_buffers = Any()
@@ -88,6 +111,13 @@ class MappingKernelManager(MultiKernelManager):
     def _default_kernel_buffers(self):
         return defaultdict(lambda: {'buffer': [], 'session_key': '', 'channels': {}})
 
+    last_kernel_activity = Instance(datetime,
+        help="The last activity on any kernel, including shutting down a kernel")
+
+    def __init__(self, **kwargs):
+        super(MappingKernelManager, self).__init__(**kwargs)
+        self.last_kernel_activity = utcnow()
+
     #-------------------------------------------------------------------------
     # Methods for managing kernels and sessions
     #-------------------------------------------------------------------------
@@ -105,7 +135,7 @@ class MappingKernelManager(MultiKernelManager):
         while not os.path.isdir(os_path) and os_path != self.root_dir:
             os_path = os.path.dirname(os_path)
         return os_path
-    
+
     @gen.coroutine
     def start_kernel(self, kernel_id=None, path=None, **kwargs):
         """Start a kernel for a session and return its kernel_id.
@@ -148,7 +178,7 @@ class MappingKernelManager(MultiKernelManager):
 
         # py2-compat
         raise gen.Return(kernel_id)
-    
+
     def start_buffering(self, kernel_id, session_key, channels):
         """Start buffering messages for a kernel
 
@@ -163,6 +193,12 @@ class MappingKernelManager(MultiKernelManager):
         channels: dict({'channel': ZMQStream})
             The zmq channels whose messages should be buffered.
         """
+
+        if not self.buffer_offline_messages:
+            for channel, stream in channels.items():
+                stream.close()
+            return
+
         self.log.info("Starting buffering for %s", session_key)
         self._check_kernel_id(kernel_id)
         # clear previous buffering state
@@ -182,7 +218,6 @@ class MappingKernelManager(MultiKernelManager):
         for channel, stream in channels.items():
             stream.on_recv(partial(buffer_msg, channel))
 
-    
     def get_buffer(self, kernel_id, session_key):
         """Get the buffer for a given kernel
 
@@ -226,7 +261,6 @@ class MappingKernelManager(MultiKernelManager):
         for stream in buffer_info['channels'].values():
             if not stream.closed():
                 stream.on_recv(None)
-                stream.socket.close()
                 stream.close()
 
         msg_buffer = buffer_info['buffer']
@@ -239,8 +273,10 @@ class MappingKernelManager(MultiKernelManager):
         self._check_kernel_id(kernel_id)
         kernel = self._kernels[kernel_id]
         kernel._activity_stream.close()
+        kernel._activity_stream = None
         self.stop_buffering(kernel_id)
         self._kernel_connections.pop(kernel_id, None)
+        self.last_kernel_activity = utcnow()
         return super(MappingKernelManager, self).shutdown_kernel(kernel_id, now=now)
 
     def restart_kernel(self, kernel_id):
@@ -281,7 +317,7 @@ class MappingKernelManager(MultiKernelManager):
         kernel.session.send(channel, "kernel_info_request")
         channel.on_recv(on_reply)
         loop = IOLoop.current()
-        timeout = loop.add_timeout(loop.time() + 30, on_timeout)
+        timeout = loop.add_timeout(loop.time() + self.kernel_info_timeout, on_timeout)
         return future
 
     def notify_connect(self, kernel_id):
@@ -346,7 +382,7 @@ class MappingKernelManager(MultiKernelManager):
 
         def record_activity(msg_list):
             """Record an IOPub message arriving from a kernel"""
-            kernel.last_activity = utcnow()
+            self.last_kernel_activity = kernel.last_activity = utcnow()
 
             idents, fed_msg_list = session.feed_identities(msg_list)
             msg = session.deserialize(fed_msg_list)
@@ -365,17 +401,13 @@ class MappingKernelManager(MultiKernelManager):
         """
         if not self._initialized_culler and self.cull_idle_timeout > 0:
             if self._culler_callback is None:
-                if self.cull_idle_timeout < self.cull_idle_timeout_minimum:
-                    self.log.warning("'cull_idle_timeout' (%s) is less than the minimum value (%s) and has been set to the minimum.",
-                        self.cull_idle_timeout, self.cull_idle_timeout_minimum)
-                    self.cull_idle_timeout = self.cull_idle_timeout_minimum
                 loop = IOLoop.current()
                 if self.cull_interval <= 0: #handle case where user set invalid value
                     self.log.warning("Invalid value for 'cull_interval' detected (%s) - using default value (%s).",
                         self.cull_interval, self.cull_interval_default)
                     self.cull_interval = self.cull_interval_default
                 self._culler_callback = PeriodicCallback(
-                    self.cull_kernels, 1000*self.cull_interval, loop)
+                    self.cull_kernels, 1000*self.cull_interval)
                 self.log.info("Culling kernels with idle durations > %s seconds at %s second intervals ...",
                     self.cull_idle_timeout, self.cull_interval)
                 if self.cull_busy:
@@ -414,4 +446,3 @@ class MappingKernelManager(MultiKernelManager):
                 self.log.warning("Culling '%s' kernel '%s' (%s) with %d connections due to %s seconds of inactivity.",
                                  kernel.execution_state, kernel.kernel_name, kernel_id, connections, idle_duration)
                 self.shutdown_kernel(kernel_id)
-
